import random

import numpy as np
import torch

import hyclib as lib

def test_set_seed():
    a0 = random.random()
    b0 = np.random.random()
    c0 = torch.rand(1)
    
    with lib.random.set_seed(0):
        a1 = random.random()
        b1 = np.random.random()
        c1 = torch.rand(1)

    with lib.random.set_seed(1):
        a2 = random.random()
        b2 = np.random.random()
        c2 = torch.rand(1)
        
    a3 = random.random()
    b3 = np.random.random()
    c3 = torch.rand(1)
    
    with lib.random.set_seed(0):
        a4 = random.random()
        b4 = np.random.random()
        c4 = torch.rand(1)

    with lib.random.set_seed(1):
        a5 = random.random()
        b5 = np.random.random()
        c5 = torch.rand(1)
        
    assert a1 == a4 and b1 == b4 and c1 == c4 # seed 0 vs seed 0
    assert a2 == a5 and b2 == b5 and c2 == c5 # seed 1 vs seed 1
    assert a1 != a2 and b1 != b2 and c1 != c2 # seed 0 vs seed 1
    assert a0 != a1 and b0 != b1 and c0 != c1 # random state vs seed 0
    assert a0 != a2 and b0 != b2 and c0 != c2 # random state vs seed 1
    assert a0 != a3 and b0 != b3 and c0 != c3 # random state vs random state


import numpy as np

from ..core.itertools import flatten_seq

__all__ = [
    'isconst',
    'meshgrid_dd',
    'meshgrid',
    'inv_perm',
    'unique_rows',
    'intersect_rows',
    'repeat',
]

def isconst(x, axis=None, **kwargs):
    x = np.asanyarray(x)
    
    if axis is None:
        x = x.reshape(-1)
    else:
        if isinstance(axis, int):
            axis = [axis]
        axis = sorted([d % x.ndim for d in axis])[::-1]
        for d in axis:
            x = np.moveaxis(x, d,-1)
        x = x.reshape(*x.shape[:-len(axis)],-1)
        
    if isinstance(x, np.floating):
        return np.isclose(x[...,:-1], x[...,1:], **kwargs).all(axis=-1)
    return (x[...,:-1] == x[...,1:]).all(axis=-1)

def meshgrid_dd(*arrs):
    """
    Generalized np.meshgrid
    Mesh together list of arrays of shapes (n_1_1,...,n_1_{M_1},N_1), (n_2_1,...,n_2_{M_2},N_2), ..., (n_P_1, ..., n_P_{M_P},N_P)
    Returns arrays of shapes 
    (n_1_1,...,n_1_{M_1},n_2_1,...,n_2_{M_2},...,n_P_1, ..., n_P_{M_P},N_1),
    (n_1_1,...,n_1_{M_1},n_2_1,...,n_2_{M_2},...,n_P_1, ..., n_P_{M_P},N_2),
    ...
    (n_1_1,...,n_1_{M_1},n_2_1,...,n_2_{M_2},...,n_P_1, ..., n_P_{M_P},N_P)
    
    IMPORTANT: Data is NOT copied, unlike numpy which copies by default
    """
    sizes = [list(arr.shape[:-1]) for arr in arrs] # [[n_1,...,n_{M_1}],[n_1,...,.n_{M_2}],...]
    Ms = np.array([arr.ndim - 1 for arr in arrs]) # [M_1, M_2, ...]
    M_befores = np.cumsum(np.insert(Ms[:-1],0,0))
    M_afters = np.sum(Ms) - np.cumsum(Ms)
    Ns = [arr.shape[-1] for arr in arrs]
    shapes = [[1]*M_befores[i]+sizes[i]+[1]*M_afters[i]+[Ns[i]] for i, arr in enumerate(arrs)]
    expanded_arrs = [np.broadcast_to(arr.reshape(shapes[i]), flatten_seq(sizes)+[Ns[i]]) for i, arr in enumerate(arrs)]
    return expanded_arrs

def meshgrid(*arrs, **kwargs):
    """
    Generalized np.meshgrid
    Mesh together list of arrays of shapes (n_1_1,...,n_1_{M_1}), (n_2_1,...,n_2_{M_2}), ..., (n_P_1, ..., n_P_{M_P})
    Returns arrays of shapes 
    (n_1_1,...,n_1_{M_1},n_2_1,...,n_2_{M_2},...,n_P_1, ..., n_P_{M_P}),
    (n_1_1,...,n_1_{M_1},n_2_1,...,n_2_{M_2},...,n_P_1, ..., n_P_{M_P}),
    ...
    (n_1_1,...,n_1_{M_1},n_2_1,...,n_2_{M_2},...,n_P_1, ..., n_P_{M_P})
    
    IMPORTANT: By default, indexing='ij' rather than 'xy' as in np.meshgrid. This follows the pytorch convention.
    IMPORTANT: Data is NOT copied, unlike numpy which copies by default
    """
    default_kwargs = {
        'indexing': 'ij',
    }
    kwargs = default_kwargs | kwargs
    invalid_keys = set(kwargs.keys()) - {'indexing'}
    if len(invalid_keys) > 0:
        raise TypeError(f"meshgrid() got an unexpected keyword argument '{invalid_keys.pop()}'")
    indexing = kwargs['indexing']    
        
    arrs = (arr[...,None] for arr in arrs)
    arrs = meshgrid_dd(*arrs)
    arrs = [arr.squeeze(-1) for arr in arrs]
    
    if indexing == 'xy':
        arrs = [np.swapaxes(arr, 0, 1) for arr in arrs]
        
    return arrs

def inv_perm(p):
    """
    Code taken from: https://stackoverflow.com/questions/11649577/how-to-invert-a-permutation-array-in-numpy
    Return an array s with which np.array_equal(arr[p][s], arr) is True.
    The array_like argument p must be some permutation of 0, 1, ..., len(p)-1.
    """
    p = np.asanyarray(p) # in case p is a tuple, etc.
    s = np.empty_like(p)
    s[p] = np.arange(p.size)
    return s

def unique_rows(arr, sorted=True, return_index=False, return_inverse=False, return_counts=False):
    """
    Code modified from https://github.com/scikit-image/scikit-image/blob/main/skimage/util/unique.py
    Much faster than np.unique(arr, axis=0)
    See https://github.com/numpy/numpy/issues/11136
    Set sorted=False for maximum speed.
    In this implementation, equal_nan=True.
    
    For speed comparison:
    N, M = 100000,100
    arr = np.random.randint(M, size=(N,2)).astype(float)
    
    %timeit unique_rows(arr, sorted=False, return_index=True, return_inverse=True)
    25.9 ms ± 386 µs per loop (mean ± std. dev. of 7 runs, 10 loops each)
    
    %timeit np.unique(arr, return_index=True, return_inverse=True, axis=0)
    85.8 ms ± 240 µs per loop (mean ± std. dev. of 7 runs, 10 loops each)
    """
    if arr.ndim != 2:
        raise ValueError("unique_rows() only makes sense for 2D arrays, "
                         f"got {arr.ndim}")
    # the view in the next line only works if the array is C-contiguous
    arr = np.ascontiguousarray(arr)
    # np.unique() finds identical items in a raveled array. To make it
    # see each row as a single item, we create a view of each row as a
    # byte string of length itemsize times number of columns in `ar`
    arr_row_view = arr.view(f"|S{arr.itemsize * arr.shape[1]}")
    out = np.unique(arr_row_view, return_index=True, return_inverse=return_inverse, return_counts=return_counts)
    out = list(out)
    out[0] = arr[out[1]]
    
    if sorted:
        idx = np.lexsort(out[0].T[::-1])
        out[0] = out[0][idx]
        if return_index:
            out[1] = out[1][idx]
        if return_inverse:
            out[2] = inv_perm(idx)[out[2]]
        if return_counts:
            i = 3 if return_inverse else 2
            out[i] = out[i][idx]
    
    if not return_index:
        out.pop(1)

    if len(out) == 1:
        return out[0]
    return out

def intersect_rows(larr, rarr, return_indices=False, **kwargs):
    """
    intersect1d but for 2D arrays. Uses the same trick as unique_rows.
    """
    if larr.ndim != 2 or rarr.ndim != 2:
        raise ValueError("intersect_rows() only makes sense for 2D arrays, "
                         f"got {larr.ndim=}, {rarr.ndim=}")
    if larr.shape[1] != rarr.shape[1]:
        raise ValueError(f"larr and rarr must both have same number of columns, but {larr.shape[1]=}, {rarr.shape[1]=}.")
        
    N_cols = larr.shape[1]
    larr, rarr = np.ascontiguousarray(larr), np.ascontiguousarray(rarr)
    larr_row_view, rarr_row_view = larr.view(f"|S{larr.itemsize * N_cols}"), rarr.view(f"|S{rarr.itemsize * N_cols}")
    
    out = np.intersect1d(larr_row_view, rarr_row_view, return_indices=True, **kwargs)
    out = list(out)
    out[0] = larr[out[1]]
    
    if not return_indices:
        return out[0]
    return out
    
def repeat(arr, repeats, chunks=None):
    """
    Generalized np.repeat
    Copied from @MadPhysicist's solution: https://stackoverflow.com/questions/63510977/repeat-but-in-variable-sized-chunks-in-numpy
    
    Example due to @MadPhysicist in the same link:
    arr = np.array([0, 1, 2, 10, 11, 20, 21, 22, 23])
    #               >     <  >    <  >            <
    chunks = np.array([3, 2, 4])
    repeats = np.array([1, 3, 2])
    
    print(repeat(arr, repeats, chunks=chunks))
    >>>      [0, 1, 2, 10, 11, 10, 11, 10, 11, 20, 21, 22, 23, 20, 21, 22, 23])
    # repeats:>  1  <  >         3          <  >              2             <
    """
    if chunks is None:
        return np.repeat(arr, repeats)
    
    arr, repeats, chunks = np.asanyarray(arr), np.asanyarray(repeats), np.asanyarray(chunks)
    
    if arr.ndim != 1 or repeats.ndim != 1 or chunks.ndim != 1:
        raise ValueError(f"arr, repeats, and chunks must all be 1D, but {arr.ndim=}, {repeats.ndim=} and {chunks.ndim=}.")
    if len(repeats) != len(chunks):
        raise ValueError(f"repeats and chunks must have the same length, but {len(repeats)=} and {len(chunks)=}.")
    if chunks.sum() != len(arr):
        raise ValueError(f"sum of chunks must be the length of arr, but {chunks.sum()=} and {len(arr)=}.")

    regions = chunks * repeats
    index = np.arange(regions.sum())

    segments = np.repeat(chunks, repeats)
    resets = np.cumsum(segments[:-1])
    offsets = np.zeros_like(index)
    offsets[resets] = segments[:-1]
    offsets[np.cumsum(regions[:-1])] -= chunks[:-1]

    index -= np.cumsum(offsets)

    out = arr[index]
    
    return out



import reprlib
import functools
import warnings

class rpartial:
    """New function with partial application of the given arguments
    and keywords, but the partial args are appended to the function 
    args rather than prepended as in functools.partial.
    
    Implementation adapted from the cpython 3.11 source code for
    functools.partial.
    """

    __slots__ = "func", "args", "keywords", "__dict__", "__weakref__"

    def __new__(cls, func, /, *args, **keywords):
        if not callable(func):
            raise TypeError("the first argument must be callable")

        if hasattr(func, "func"):
            args = args + func.args
            keywords = {**func.keywords, **keywords}
            func = func.func

        self = super(rpartial, cls).__new__(cls)

        self.func = func
        self.args = args
        self.keywords = keywords
        return self

    def __call__(self, /, *args, **keywords):
        keywords = {**self.keywords, **keywords}
        return self.func(*args, *self.args, **keywords)

    @reprlib.recursive_repr()
    def __repr__(self):
        qualname = type(self).__qualname__
        args = [repr(self.func)]
        args.extend(repr(x) for x in self.args)
        args.extend(f"{k}={v!r}" for (k, v) in self.keywords.items())
        return f"{qualname}({', '.join(args)})"

    def __reduce__(self):
        return type(self), (self.func,), (self.func, self.args,
               self.keywords or None, self.__dict__ or None)

    def __setstate__(self, state):
        if not isinstance(state, tuple):
            raise TypeError("argument to __setstate__ must be a tuple")
        if len(state) != 4:
            raise TypeError(f"expected 4 items in state, got {len(state)}")
        func, args, kwds, namespace = state
        if (not callable(func) or not isinstance(args, tuple) or
           (kwds is not None and not isinstance(kwds, dict)) or
           (namespace is not None and not isinstance(namespace, dict))):
            raise TypeError("invalid partial state")

        args = tuple(args) # just in case it's a subclass
        if kwds is None:
            kwds = {}
        elif type(kwds) is not dict: # XXX does it need to be *exactly* dict?
            kwds = dict(kwds)
        if namespace is None:
            namespace = {}

        self.__dict__ = namespace
        self.func = func
        self.args = args
        self.keywords = kwds

def parametrized(dec):
    """
    A decorator for decorators.
    Copied from @Dacav's answer here: https://stackoverflow.com/questions/5929107/decorators-with-parameters
    """
    def layer(*args, **kwargs):
        def repl(f):
            return dec(f, *args, **kwargs)
        return repl
    return layer

@parametrized
def deprecated(func, msg=""):
    """This is a decorator which can be used to mark functions
    as deprecated."""
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.warn(f"Call to deprecated function {func.__name__}. {msg}",
                      category=DeprecationWarning,
                      stacklevel=2)
        return func(*args, **kwargs)
    return new_func



import pytest

import hyclib as lib

def bad_iterable():
    yield from [1/0,2,2]

@pytest.mark.parametrize('iterable, expected', [
    (range(0), True),
    (range(1), True),
    (range(2), False),
    ([1,2,1], False),
    ([2,2,2], True),
    (bad_iterable(), ZeroDivisionError),
])
def test_isconst(iterable, expected):
    if not isinstance(expected, bool):
        with pytest.raises(expected):
            lib.itertools.isconst(iterable)
    else:
        assert lib.itertools.isconst(iterable) == expected
        
@pytest.mark.parametrize('iterable, expected', [
    (range(0), True),
    (range(1), False),
    (range(2), False),
    ([], True),
    ([1,2,1], False),
    (bad_iterable(), ZeroDivisionError),
])
def test_isempty(iterable, expected):
    if not isinstance(expected, bool):
        with pytest.raises(expected):
            lib.itertools.isempty(iterable)
    else:
        assert lib.itertools.isempty(iterable) == expected

@pytest.mark.parametrize('s, depth, expected', [
    ([[1,2,[3]],[],(4,5),6], 1, [1,2,[3],4,5,6]),
    ([[1,2,[3]],[],(4,5),6], -1, [1,2,3,4,5,6]),
    (tuple([[1,2,[3]],[],(4,5),6]), -1, (1,2,3,4,5,6)),
])
def test_flatten_seq(s, depth, expected):
    assert str(lib.itertools.flatten_seq(s, depth=depth)) == str(expected)
    
@pytest.fixture
def iterable():
    return [[1,2,[3]],[],(4,5),6]
    
@pytest.mark.parametrize('depth, types, exclude, expected', [
    (0, None, None, [[1,2,[3]],[],(4,5),6]),
    (1, None, None, [1,2,[3],4,5,6]),
    (-1, None, None, [1,2,3,4,5,6]),
    (0, [], None, [[1,2,[3]],[],(4,5),6]),
    (1, [], None, [[1,2,[3]],[],(4,5),6]),
    (-1, [], None, [[1,2,[3]],[],(4,5),6]),
    (0, [list], None, [[1,2,[3]],[],(4,5),6]),
    (1, [list], None, [1,2,[3],(4,5),6]),
    (-1, [list], None, [1,2,3,(4,5),6]),
    (0, None, [list], [[1,2,[3]],[],(4,5),6]),
    (1, None, [list], [[1,2,[3]],[],4,5,6]),
    (-1, None, [list], [[1,2,[3]],[],4,5,6]),
    (0, [], [list], [[1,2,[3]],[],(4,5),6]),
    (1, [], [list], [[1,2,[3]],[],(4,5),6]),
    (-1, [], [list], [[1,2,[3]],[],(4,5),6]),
    (0, [list], [list], [[1,2,[3]],[],(4,5),6]),
    (1, [list], [list], [[1,2,[3]],[],(4,5),6]),
    (-1, [list], [list], [[1,2,[3]],[],(4,5),6]),
    (0, None, [tuple], [[1,2,[3]],[],(4,5),6]),
    (1, None, [tuple], [1,2,[3],(4,5),6]),
    (-1, None, [tuple], [1,2,3,(4,5),6]),
])
def test_deep_iter(iterable, depth, types, exclude, expected):
    for x, y in zip(lib.itertools.deep_iter(iterable, depth=depth, types=types, exclude=exclude), expected):
        assert x == y
    
@pytest.mark.parametrize('d, depth, expected', [
    ({'a': {'b': {'c': 'd'}}}, 1, {'a.b': {'c': 'd'}}),
    ({'a': {'b': {'c': 'd'}}}, -1, {'a.b.c': 'd'}),
    ({'abc': {'bcd': {'cde': 'def'}}}, -1, {'abc.bcd.cde': 'def'}),
])
def test_flatten_dict(d, depth, expected):
    assert str(lib.itertools.flatten_dict(d, depth=depth)) == str(expected)
    
@pytest.fixture
def d():
    return {
        'a': 'hi',
        'b': {
            'c': 1,
            'd': 1.0,
            'e': {}
        },
        'f': 'hii'
    }
    
@pytest.mark.parametrize('delimiter, expected', [
    ('.', {'a': 'hi', 'b.c': 1, 'b.d': 1.0, 'f': 'hii'}),
    ('/', {'a': 'hi', 'b/c': 1, 'b/d': 1.0, 'f': 'hii'}),
])
def test_dict_iter(d, delimiter, expected):
    for x, y in zip(lib.itertools.dict_iter(d, delimiter=delimiter), expected.items()):
        assert x == y
    
@pytest.mark.parametrize('d, depth, expected', [
    ({'a.b': {'b': {'c': 'd'}}}, 1, pytest.raises(ValueError)),
    ({'a.b': {'b': {'c': 'd'}}}, -1, pytest.raises(ValueError)),
])
def test_flatten_dict_exceptions(d, depth, expected):
    with expected:
        lib.itertools.flatten_dict(d, depth=depth)
    
@pytest.mark.parametrize('d, keys, value, expected', [
    ({}, ['a', 'b'], 'c', {'a': {'b': 'c'}}),
])
def test_assign_dict(d, keys, value, expected):
    lib.itertools.assign_dict(d, keys, value)
    assert str(d) == str(expected)
    
@pytest.mark.parametrize('enum', [
    False,
    True
])
def test_product(enum):
    l1 = [
        'hi',
        'bye',
    ]
    l2 = [
        3,
        7,
        9,
    ]
    expected_ndindices = [
        (0,0),
        (0,1),
        (0,2),
        (1,0),
        (1,1),
        (1,2),
    ]
    for i, elem in enumerate(lib.itertools.product(l1, l2, enum=enum)):
        if enum:
            ndidx, (e1, e2) = elem
            assert ndidx == expected_ndindices[i]
        else:
            e1, e2 = elem
        assert e1 == l1[expected_ndindices[i][0]]
        assert e2 == l2[expected_ndindices[i][1]]
        
@pytest.mark.parametrize('d1, d2, mode, fillvalue, expected', [
    ({'a': 1, 'b': 2}, {'b': 3, 'a': 4}, 'strict', None, {'a': (1,4), 'b': (2,3)}),
    ({'c': 5, 'a': 1, 'b': 2}, {'b': 3, 'a': 4, 'd': 6}, 'intersect', None, {'a': (1,4), 'b': (2,3)}),
    ({'c': 5, 'a': 1, 'b': 2}, {'d': 6, 'e': 7}, 'intersect', None, {}),
    ({'c': 5, 'a': 1, 'b': 2}, {'b': 3, 'a': 4, 'd': 6}, 'union', None, {'a': (1,4), 'b': (2,3), 'c': (5,None), 'd': (None,6)}),
    ({'c': 5, 'a': 1, 'b': 2}, {'b': 3, 'a': 4, 'd': 6}, 'union', 10, {'a': (1,4), 'b': (2,3), 'c': (5,10), 'd': (10,6)}),
])
def test_dict_zip(d1, d2, mode, fillvalue, expected):
    if mode != 'union':
        d = {k: (v1, v2) for k, v1, v2 in lib.itertools.dict_zip(d1, d2, mode=mode)}
    else:
        d = {k: (v1, v2) for k, v1, v2 in lib.itertools.dict_zip(d1, d2, mode=mode, fillvalue=fillvalue)}
    assert d == expected # no need for order of elemenets to be equal
    
@pytest.mark.parametrize('d1, d2, mode, expected', [
    ({'a': 1, 'b': 2, 'c': 3}, {'b': 3, 'a': 4}, 'strict', pytest.raises(ValueError)),
    ({'a': 1, 'c': 5}, {'b': 3, 'a': 4}, 'strict', pytest.raises(KeyError)),
    ({'a': 1, 'b': 2}, {'c': 5, 'b': 3, 'a': 4}, 'strict', pytest.raises(ValueError)),
    ({'a': 1, 'b': 2}, {'b': 3, 'a': 4}, 'whatever', pytest.raises(NotImplementedError)),
])
def test_dict_zip_exceptions(d1, d2, mode, expected):
    with expected:
        list(lib.itertools.dict_zip(d1, d2, mode=mode))

def test_dict_union():
    d = {
        'options': {
            'max_num_steps': 100,
        }
    }
    new_d = {
        'options': {
            'dtype': 'float32',
        }
    }
    expected = {
        'options': {
            'max_num_steps': 100,
            'dtype': 'float32',
        }
    }
    assert str(lib.itertools.dict_union(d, new_d)) == str(expected)


import pytest

import hyclib as lib

def test_logging():
    # just ensure there is no error, I don't know how to write tests for this
    parser = lib.logging.basic_parser()
    args = parser.parse_args([])
    lib.logging.basic_config(**vars(args))


import numpy as np
from numpy.polynomial import Polynomial
import pytest

import hyclib as lib

def check_npf_array(arr, expected_func_arr):
    expected_func_arr = np.array(expected_func_arr)
    expected_shape = expected_func_arr.shape
    assert isinstance(arr, lib.npf.ndarray)
    assert arr.shape == expected_shape
    xs = np.array([-2.0,-1.0,0.0,1.0,2.0,-2.0j,-1.0j,1.0j,2.0j,2.0-2.0j,2.0-1.0j,2.0+0.0j,2.0+1.0j,2.0+2.0j])
    for x in xs:
        res = arr(x)
        assert res.shape == expected_shape
        assert np.allclose(res.reshape(-1), np.array([func(x) for func in expected_func_arr.reshape(-1)]))
    res = arr(xs)
    assert res.shape == (*xs.shape, *expected_shape)
    assert np.allclose(res.reshape(-1), np.array([[func(x) for func in expected_func_arr.reshape(-1)] for x in xs]).reshape(-1))
    xs = xs.reshape(2,7)
    assert arr(xs).shape == (*xs.shape, *expected_shape)
    
def test_shape():
    f1, f2 = lambda x: 2*x, lambda x: 3*x
    a = lib.npf.array([[f1, f2]])
    x = np.ones((3,4))
    assert a(x).shape == (3,4,1,2)
    assert a(x, batch='trailing').shape == (1,2,3,4)

def test():
    f1, f2 = lambda x: 2*x, lambda x: 3*x
    g1, g2 = lambda x: 4*x, lambda x: 5*x
    a = lib.npf.array([[f1, f2]])
    b = lib.npf.array([g1, g2])
    
    # check shape
    # print(a, b)
    check_npf_array(a, [[f1, f2]])
    check_npf_array(b, [g1, g2])
    
    # check add two functions
    c = a+b
    # print(c)
    check_npf_array(c, [[lambda x: f1(x)+g1(x), lambda x: f2(x)+g2(x)]])
    
    # check slicing
    d = c[:,:1]
    # print(d)
    check_npf_array(d, [[lambda x: f1(x)+g1(x)]])
    
    # check that new function does not retain reference to old function
    a[0] = lambda x: x
    b[0] = lambda x: x
    # print(a, b)
    check_npf_array(a, [[lambda x: x, lambda x: x]])
    check_npf_array(b, [lambda x: x, g2])
    check_npf_array(a+b, [[lambda x: x+x, lambda x: x+g2(x)]])
    check_npf_array(c, [[lambda x: f1(x)+g1(x), lambda x: f2(x)+g2(x)]]) # should be invariant to assignments to a and b
    check_npf_array(d, [[lambda x: f1(x)+g1(x)]]) # should be invariant to assignments to a and b
    
    a = lib.npf.array([[f1, f2]])
    b = lib.npf.array([g1, g2])
    
    # check a bunch of basic operators
    c = b+a
    # print(c)
    check_npf_array(c, [[lambda x: f1(x)+g1(x), lambda x: f2(x)+g2(x)]])
    
    c = a-b
    # print(c)
    check_npf_array(c, [[lambda x: f1(x)-g1(x), lambda x: f2(x)-g2(x)]])
    
    c = a*b
    # print(c)
    check_npf_array(c, [[lambda x: f1(x)*g1(x), lambda x: f2(x)*g2(x)]])
    
    g1, g2 = lambda x: 4*x+3, lambda x: 5*x+4
    b = lib.npf.array([g1, g2])
    c = a/b
    # print(c)
    check_npf_array(c, [[lambda x: f1(x)/g1(x), lambda x: f2(x)/g2(x)]])
    
    c = a**b
    # print(c)
    check_npf_array(c, [[lambda x: f1(x)**g1(x), lambda x: f2(x)**g2(x)]])
    
#     # check operations respect parantheses
#     c = b**((a+b)*b-a)
#     print(c)
    
    # check that operations work with scalars
    b = np.pi
    check_npf_array(a+b, [[lambda x: f1(x)+b, lambda x: f2(x)+b]]) # scalar + npf_func
    check_npf_array(b+a, [[lambda x: f1(x)+b, lambda x: f2(x)+b]]) # npf_func + scalar
    
    b = np.array([np.pi, 2*np.pi])
    check_npf_array(a+b, [[lambda x: f1(x)+b[0], lambda x: f2(x)+b[1]]]) # npf_func + array_scalar
    check_npf_array(b+a, [[lambda x: f1(x)+b[0], lambda x: f2(x)+b[1]]]) # array_scalar + npf_func
    
    b = lib.npf.array([np.pi, 2*np.pi])
    check_npf_array(a+b, [[lambda x: f1(x)+b[0], lambda x: f2(x)+b[1]]]) # npf_func + npf_scalar
    check_npf_array(b+a, [[lambda x: f1(x)+b[0], lambda x: f2(x)+b[1]]]) # npf_scalar + npf_func
    
    a = lib.npf.array([[1.0,2.0]])
    check_npf_array(a+b, [[lambda x: a[0,0]+b[0], lambda x: a[0,1]+b[1]]]) # npf_scalar + array_scalar
    check_npf_array(b+a, [[lambda x: a[0,0]+b[0], lambda x: a[0,1]+b[1]]]) # array_scalar + npf_scalar
    
    # check that operations work with polynomials
    a = lib.npf.array([[Polynomial([3,2,1]), Polynomial([1,2,3])]])
    b = lib.npf.array([Polynomial([4,5]), Polynomial([5,4])])
    check_npf_array(a+b, [[a[0,0]+b[0], a[0,1]+b[1]]]) # npf_poly + npf_poly
    
    b = lib.npf.array([g1, g2])
    check_npf_array(a+b, [[lambda x: a[0,0](x)+g1(x), lambda x: a[0,1](x)+g2(x)]]) # npf_poly + npf_func
    check_npf_array(b+a, [[lambda x: a[0,0](x)+g1(x), lambda x: a[0,1](x)+g2(x)]]) # npf_func + npf_poly
    
    # b = Polynomial([1,2])
    # check_npf_array(a+b, [[lambda x: a[0,0](x)+b(x), lambda x: a[0,1](x)+b(x)]]) # npf_poly + poly
    # check_npf_array(b+a, [[lambda x: a[0,0](x)+b(x), lambda x: a[0,1](x)+b(x)]]) # poly + npf_poly
    
    b = lib.npf.array([np.pi, 2*np.pi])
    check_npf_array(a+b, [[a[0,0]+b[0], a[0,1]+b[1]]]) # npf_poly + npf_scalar
    check_npf_array(b+a, [[a[0,0]+b[0], a[0,1]+b[1]]]) # npf_scalar + npf_poly
    
    b = np.array([np.pi, 2*np.pi])
    check_npf_array(a+b, [[a[0,0]+b[0], a[0,1]+b[1]]]) # npf_poly + array_scalar
    check_npf_array(b+a, [[a[0,0]+b[0], a[0,1]+b[1]]]) # array_scalar + npf_poly
    
    b = np.pi
    check_npf_array(a+b, [[a[0,0]+b, a[0,1]+b]]) # npf_poly + scalar
    check_npf_array(b+a, [[a[0,0]+b, a[0,1]+b]]) # scalar + npf_poly
    
@pytest.mark.parametrize('a, b, inv', [
    (lib.npf.array([[lambda x: 2*x, lambda x: 3*x],[lambda x: 4*x, lambda x: 5*x]]),
     lib.npf.array([[lambda x: 6*x, lambda x: 7*x],[lambda x: 8*x, lambda x: 9*x]]), True),
    (lib.npf.array([[Polynomial([1,2,3]), Polynomial([4,5])],[Polynomial([2,1]), Polynomial([6,5,4])]]),
     lib.npf.array([[Polynomial([2,3]), Polynomial([5,6,7])],[Polynomial([3,2]), Polynomial([5,4,3])]]), False),
    (lib.npf.array([[1,2],[3.0,5.0]]),
     lib.npf.array([[3,4],[5.0,6.0]]), True),
])
def test_linalg(a, b, inv):
    x = np.linspace(1,2,num=3)
    
    result, expected = (a @ b)(x), a(x) @ b(x)
    assert (result.shape == expected.shape) and np.allclose(result, expected)
    
    result, expected = lib.npf.trace(a)(x), np.trace(a(x), axis1=-2, axis2=-1)
    assert (result.shape == expected.shape) and np.allclose(result, expected)
    
    result, expected = lib.npf.det(a)(x), np.linalg.det(a(x))
    assert (result.shape == expected.shape) and np.allclose(result, expected)
    
    result, expected = lib.npf.adj(a)(x), np.linalg.inv(a(x))*np.linalg.det(a(x))[:,None,None]
    assert (result.shape == expected.shape) and np.allclose(result, expected)
    
    if inv:
        result, expected = lib.npf.inv(a)(x), np.linalg.inv(a(x))
        assert (result.shape == expected.shape) and np.allclose(result, expected)



import pytest
import argparse

import hyclib as lib

def func_0(a, b: int, hi_bye: str = 'hibye', hell_o: float = 0.3, debug=False):
    return a, b, hi_bye, hell_o
    
def func_1(d, e: float, f: str = 'bye'):
    return d, e, f
   
@pytest.fixture
def parser_0():
    parser = lib.argparse.default_parser(func_0, configs={
            'a': dict(opt_str=['-a'], type=int),
            'b': dict(opt_str=['-b']),
            'hi_bye': dict(opt_str=['--hi', '--hi-bye']),
            'hell_o': dict(opt_str=['--hell', '--hell-o'], nargs='+'),
            'debug': dict(action='store_true'),
        },
        add_help=False
    )
    return parser

@pytest.fixture
def parser_1():
    parser = lib.argparse.default_parser(func_1, configs={
            'e': dict(choices=[0.5,0.7,0.9]),
        },
        add_help=False,
    )
    return parser

@pytest.mark.parametrize('args, expected_0, expected_1', [
    ('-a 1 -b 2 --hi hi --hell 0.3 0.4 -f hihi blah 0.5'.split(),
     {'a': 1, 'b': 2, 'hi_bye': 'hi', 'hell_o': [0.3,0.4], 'debug': False},
     {'d': 'blah', 'e': 0.5, 'f': 'hihi'}),
    ('blah 0.5 -a 1 -b 2 --hi hi --hell 0.3 0.4 -f hihi'.split(),
     {'a': 1, 'b': 2, 'hi_bye': 'hi', 'hell_o': [0.3,0.4], 'debug': False},
     {'d': 'blah', 'e': 0.5, 'f': 'hihi'}),
    ('-a 1 -b 2 --hi hi --hell 0.3 0.4 -f hihi blah 0.5 --debug'.split(),
     {'a': 1, 'b': 2, 'hi_bye': 'hi', 'hell_o': [0.3,0.4], 'debug': True},
     {'d': 'blah', 'e': 0.5, 'f': 'hihi'}),
    ('-a 1 -b 2 -f hihi blah 0.5 --debug'.split(),
     {'a': 1, 'b': 2, 'hi_bye': 'hibye', 'hell_o': 0.3, 'debug': True},
     {'d': 'blah', 'e': 0.5, 'f': 'hihi'}),
])
def test_multi_parse(parser_0, parser_1, args, expected_0, expected_1):
    args_0, args_1 = lib.argparse.multi_parse([parser_0, parser_1], args=args)
    assert vars(args_0) == expected_0
    assert vars(args_1) == expected_1
    
    
@pytest.mark.parametrize('args, expected_0, expected_1', [
    ('-a 1 -b 2 --hi hi --hell 0.3 0.4 -f hihi blah 0.6'.split(),
     {'a': 1, 'b': 2, 'hi_bye': 'hi', 'hell_o': [0.3,0.4], 'debug': False},
     {'d': 'blah', 'e': 0.6, 'f': 'hihi'}),
])
def test_multi_parse_bad_choice(parser_0, parser_1, args, expected_0, expected_1):
    with pytest.raises(SystemExit):
        args_0, args_1 = lib.argparse.multi_parse([parser_0, parser_1], args=args)
        
def test_multi_parse_help(parser_0, parser_1):
    with pytest.raises(SystemExit):
        lib.argparse.multi_parse([parser_0, parser_1], args=['-h'])


import itertools

from . import (
    functools,
    pprint,
)

@functools.parametrized
def attrs(cls, *args, **kwargs):
    keys = list(args) + list(kwargs.keys())
    
    if len(set(keys)) != len(keys):
        raise TypeError(f"Overlapping attributes not allowed, but {args=}, {kwargs=}.")
        
    for arg in args:
        if not arg.isidentifier():
            raise TypeError(f"all args must be a valid string identifier, but got {arg=}.")
    
    def __init__(self, *init_args, **init_kwargs):
        if hasattr(cls, "__pre_init__"):
            cls.__pre_init__(self)
        
        if len(init_args) > len(keys):
            raise TypeError(f"__init__ takes {len(args)} to {len(keys)} positional arguments but {len(init_args)} were given")
        
        kwargs.update(init_kwargs)
        
        missing_keys = []
        for key, value in itertools.zip_longest(keys, init_args): # keys is always longer or equal to init_args due to above check
            if value is None:
                if key in kwargs:
                    value = kwargs.pop(key)
                else:
                    missing_keys.append(key)
            else:
                if key in kwargs:
                    raise TypeError(f"__init__ got multiple values for argument '{key}'")
                
            self.__dict__[key] = value
            
        if (n := len(missing_keys)) > 0:
            missing_keys = pprint.pformat_english(*map(repr, missing_keys))
            raise TypeError(f"__init__ missing {n} positional argument{'' if n == 1 else 's'}: {missing_keys}")
        
        if len(kwargs) > 0:
            raise TypeError(f"__init__ got an unexpected keyword argument '{next(iter(kwargs))}'")
            
        if hasattr(cls, "__post_init__"):
            cls.__post_init__(self)

    cls.__init__ = __init__
    return cls

# def init_repr(cls):
#     cls__init__ = cls.__init__
    
#     def __init__(self, *args, **kwargs):
#         cls__init__(self, *args, **kwargs)
#         self.args = args
#         self.kwargs = kwargs
        
#     cls.__init__ = __init__
    
#     def __repr__(self):
#         args = ", ".join(map(repr, self.args))
#         kwargs = ", ".join(f'{k}={repr(v)}' for k, v in self.kwargs.items())
#         return f'{type(self).__name__}({", ".join(filter(None, [args, kwargs]))})'
    
#     cls.__repr__ = __repr__
#     return cls

def attr_repr(cls):
    def __repr__(self):
        attrs = (f'{k}={repr(v)}' for k, v in self.__dict__.items())
        return f'{type(self).__name__}({", ".join(attrs)})'
    
    cls.__repr__ = __repr__
    return cls

def attr_str(cls):
    def __str__(self):
        attrs = (f'{k}={str(v)}' for k, v in self.__dict__.items())
        return f'{type(self).__name__}({", ".join(attrs)})'
    
    cls.__str__ = __str__
    return cls


from datetime import timedelta
from string import Template

class TimeDeltaTemplate(Template):
    delimiter = "%"
    idpattern = "(?a-i:[wdHMSf])" # ASCII-only, don't ignore case

def strftime(td, fmt):
    if not isinstance(td, timedelta):
        return f'{td:{fmt}}'
    
    units = {
        'w': (timedelta(weeks=1), None),
        'd': (timedelta(days=1), None),
        'H': (timedelta(hours=1), 2),
        'M': (timedelta(minutes=1), 2),
        'S': (timedelta(seconds=1), 2),
        'f': (timedelta(microseconds=1), 6),
    }
    template = TimeDeltaTemplate(fmt)
    
    identifiers = []
    for unit in units:
        try:
            template.substitute({k: '' for k in units if k != unit})
        except KeyError:
            identifiers.append(unit)
    # identifiers = template.get_identifiers() # available only in 3.11+

    d = {}
    for unit, (unit_td, unit_padding) in units.items():
        if unit in identifiers:
            q, td = divmod(td, unit_td)
            d[unit] = str(q) if unit_padding is None else f'{q:0{unit_padding}d}'

    return template.substitute(d)


import builtins
from operator import index
from collections import namedtuple

import numpy as np
from numpy.testing import suppress_warnings
from scipy import stats as spstats
import statsmodels.api as sm

from ..np import stats as npstats

BinnedStatisticddResult = namedtuple('BinnedStatisticddResult',
                                     ('statistic', 'bin_edges',
                                      'binnumber'))
LinregressResult = namedtuple('LinregressResult', ('slope', 'intercept', 'rvalue', 'pvalue', 'stderr', 'intercept_stderr'))

def linregress(x, y, yerr=None, alternative='two-sided'):
    if alternative != 'two-sided':
        raise NotImplementedError('currently only allow alternative == "two-sided".')
        
    if yerr is not None:
        weights = 1/yerr**2
    else:
        weights = 1.0
        
    x = np.asarray(x)
    y = np.asarray(y)
    weights = np.asarray(weights)
        
    model = sm.WLS(y, sm.add_constant(x), weights=weights)
    res = model.fit()
    res = LinregressResult(
        slope=res.params[1],
        intercept=res.params[0],
        rvalue=res.rsquared**0.5,
        pvalue=res.pvalues[1],
        stderr=res.bse[1],
        intercept_stderr=res.bse[0],
    )
    return res

def _bincount(x, weights):
    if np.iscomplexobj(weights):
        a = np.bincount(x, np.real(weights))
        b = np.bincount(x, np.imag(weights))
        z = a + b*1j

    else:
        z = np.bincount(x, weights)
    return z

def _bin_edges(sample, bins=None, range=None):
    """ Create edge arrays
    """
    Dlen, Ndim = sample.shape

    nbin = np.empty(Ndim, int)    # Number of bins in each dimension
    edges = Ndim * [None]         # Bin edges for each dim (will be 2D array)
    dedges = Ndim * [None]        # Spacing between edges (will be 2D array)

    # Select range for each dimension
    # Used only if number of bins is given.
    if range is None:
        smin = np.atleast_1d(np.array(np.nanmin(sample, axis=0), float))
        smax = np.atleast_1d(np.array(np.nanmax(sample, axis=0), float))
    else:
        if len(range) != Ndim:
            raise ValueError(
                f"range given for {len(range)} dimensions; {Ndim} required")
        smin = np.empty(Ndim)
        smax = np.empty(Ndim)
        for i in builtins.range(Ndim):
            if range[i][1] < range[i][0]:
                raise ValueError(
                    "In {}range, start must be <= stop".format(
                        f"dimension {i + 1} of " if Ndim > 1 else ""))
            smin[i], smax[i] = range[i]

    # Make sure the bins have a finite width.
    for i in builtins.range(len(smin)):
        if smin[i] == smax[i]:
            smin[i] = smin[i] - .5
            smax[i] = smax[i] + .5

    # Preserve sample floating point precision in bin edges
    edges_dtype = (sample.dtype if np.issubdtype(sample.dtype, np.floating)
                   else float)

    # Create edge arrays
    for i in builtins.range(Ndim):
        if np.isscalar(bins[i]):
            nbin[i] = bins[i] + 2  # +2 for outlier bins
            edges[i] = np.linspace(smin[i], smax[i], nbin[i] - 1,
                                   dtype=edges_dtype)
        else:
            edges[i] = np.asarray(bins[i], edges_dtype)
            nbin[i] = len(edges[i]) + 1  # +1 for outlier bins
        dedges[i] = np.diff(edges[i])

    nbin = np.asarray(nbin)

    return nbin, edges, dedges


def _bin_numbers(sample, nbin, edges, dedges, expand_binnumbers=False):
    """Compute the bin number each sample falls into, in each dimension
    """
    Dlen, Ndim = sample.shape

    sampBin = [
        np.digitize(sample[:, i], edges[i])
        for i in range(Ndim)
    ]

    # Using `digitize`, values that fall on an edge are put in the right bin.
    # For the rightmost bin, we want values equal to the right
    # edge to be counted in the last bin, and not as an outlier.
    for i in range(Ndim):
        # Find the rounding precision
        dedges_min = dedges[i].min()
        if dedges_min == 0:
            raise ValueError('The smallest edge difference is numerically 0.')
        decimal = int(-np.log10(dedges_min)) + 6
        # Find which points are on the rightmost edge.
        on_edge = np.where((sample[:, i] >= edges[i][-1]) &
                           (np.around(sample[:, i], decimal) ==
                            np.around(edges[i][-1], decimal)))[0]
        # Shift these points one bin to the left.
        sampBin[i][on_edge] -= 1

    if expand_binnumbers:
        binnumbers = np.stack(sampBin)
    else:
        # Compute the sample indices in the flattened statistic matrix.
        binnumbers = np.ravel_multi_index(sampBin, nbin)

    return binnumbers


def _create_binned_data(bin_numbers, unique_bin_numbers, values, vv):
    """ Create hashmap of bin ids to values in bins
    key: bin number
    value: list of binned data
    """
    bin_map = dict()
    for i in unique_bin_numbers:
        bin_map[i] = []
    for i in builtins.range(len(bin_numbers)):
        bin_map[bin_numbers[i]].append(values[vv, i])
    return bin_map

def _calc_binned_statistic(Vdim, bin_numbers, result, values, stat_func, values_err=None):
    unique_bin_numbers = np.unique(bin_numbers)
    for vv in builtins.range(Vdim):
        bin_map = _create_binned_data(bin_numbers, unique_bin_numbers,
                                      values, vv)
        if values_err is not None:
            err_bin_map = _create_binned_data(bin_numbers, unique_bin_numbers,
                                      values_err, vv)
        for i in unique_bin_numbers:
            stat = stat_func(np.array(bin_map[i])) if values_err is None else stat_func(np.array(bin_map[i]), np.array(err_bin_map[i]))
            if np.iscomplexobj(stat) and not np.iscomplexobj(result):
                raise ValueError("The statistic function returns complex ")
            result[vv, i] = stat

def binned_statistic_dd(sample, values, values_err=None, statistic='mean',
                        bins=10, range=None, expand_binnumbers=False,
                        binned_statistic_result=None, nan_policy='raise'):
    if nan_policy not in ['raise', 'omit']:
        raise ValueError(f"nan_policy must be 'raise' or 'omit', but {nan_policy=}.")
    
    known_stats = ['mean', 'median', 'count', 'sum', 'std', 'min', 'max']
    if not callable(statistic) and statistic not in known_stats:
        raise ValueError('invalid statistic %r' % (statistic,))

    try:
        bins = index(bins)
    except TypeError:
        # bins is not an integer
        pass
    # If bins was an integer-like object, now it is an actual Python int.

    # NOTE: for _bin_edges(), see e.g. gh-11365
    if nan_policy == 'raise' and not np.isfinite(sample).all():
        raise ValueError('%r contains non-finite values.' % (sample,))

    # `Ndim` is the number of dimensions (e.g. `2` for `binned_statistic_2d`)
    # `Dlen` is the length of elements along each dimension.
    # This code is based on np.histogramdd
    try:
        # `sample` is an ND-array.
        Dlen, Ndim = sample.shape
    except (AttributeError, ValueError):
        # `sample` is a sequence of 1D arrays.
        sample = np.atleast_2d(sample).T
        Dlen, Ndim = sample.shape

    # Store initial shape of `values` to preserve it in the output
    values = np.asarray(values)
    if values_err is not None:
        values_err = np.asarray(values_err)
        try:
            values, values_err = np.broadcast_arrays(values, values_err)
        except ValueError as err:
            raise AttributeError(f'shape of values must match shape of values_err, but got {values.shape=} and {values_err.shape=}') from err
    input_shape = list(values.shape)
    # Make sure that `values` is 2D to iterate over rows
    values = np.atleast_2d(values)
    if values_err is not None:
        values_err = np.atleast_2d(values_err)
    Vdim, Vlen = values.shape

    # Make sure `values` match `sample`
    if statistic != 'count' and Vlen != Dlen:
        raise AttributeError('The number of `values` elements must match the '
                             f'length of each `sample` dimension, but got {sample.shape=} and {values.shape=}')

    try:
        M = len(bins)
        if M != Ndim:
            raise AttributeError('The dimension of bins must be equal '
                                 'to the dimension of the sample x.')
    except TypeError:
        bins = Ndim * [bins]

    if binned_statistic_result is None:
        nbin, edges, dedges = _bin_edges(sample, bins, range)
        binnumbers = _bin_numbers(sample, nbin, edges, dedges)
    else:
        edges = binned_statistic_result.bin_edges
        nbin = np.array([len(edges[i]) + 1 for i in builtins.range(Ndim)])
        # +1 for outlier bins
        dedges = [np.diff(edges[i]) for i in builtins.range(Ndim)]
        binnumbers = binned_statistic_result.binnumber

    # Avoid overflow with double precision. Complex `values` -> `complex128`.
    if values_err is None:
        result_type = np.result_type(values, np.float64)
    else:
        result_type = np.result_type(values, values_err, np.float64)
    
    result = np.empty([Vdim, nbin.prod()], dtype=result_type)

    if statistic in {'mean', np.mean}:
        result.fill(np.nan)
        flatcount = _bincount(binnumbers, None)
        a = flatcount.nonzero()
        for vv in builtins.range(Vdim):
            flatsum = _bincount(binnumbers, values[vv])
            result[vv, a] = flatsum[a] / flatcount[a]
    elif statistic in {'std', np.std}:
        result.fill(np.nan)
        flatcount = _bincount(binnumbers, None)
        a = flatcount.nonzero()
        for vv in builtins.range(Vdim):
            flatsum = _bincount(binnumbers, values[vv])
            delta = values[vv] - flatsum[binnumbers] / flatcount[binnumbers]
            std = np.sqrt(
                _bincount(binnumbers, delta*np.conj(delta))[a] / flatcount[a]
            )
            result[vv, a] = std
        result = np.real(result)
    elif statistic == 'count':
        result = np.empty([Vdim, nbin.prod()], dtype=np.float64)
        result.fill(0)
        flatcount = _bincount(binnumbers, None)
        a = np.arange(len(flatcount))
        result[:, a] = flatcount[np.newaxis, :]
    elif statistic in {'sum', np.sum}:
        result.fill(0)
        for vv in builtins.range(Vdim):
            flatsum = _bincount(binnumbers, values[vv])
            a = np.arange(len(flatsum))
            result[vv, a] = flatsum
    elif statistic in {'median', np.median}:
        result.fill(np.nan)
        for vv in builtins.range(Vdim):
            i = np.lexsort((values[vv], binnumbers))
            _, j, counts = np.unique(binnumbers[i],
                                     return_index=True, return_counts=True)
            mid = j + (counts - 1) / 2
            mid_a = values[vv, i][np.floor(mid).astype(int)]
            mid_b = values[vv, i][np.ceil(mid).astype(int)]
            medians = (mid_a + mid_b) / 2
            result[vv, binnumbers[i][j]] = medians
    elif statistic in {'min', np.min}:
        result.fill(np.nan)
        for vv in builtins.range(Vdim):
            i = np.argsort(values[vv])[::-1]  # Reversed so the min is last
            result[vv, binnumbers[i]] = values[vv, i]
    elif statistic in {'max', np.max}:
        result.fill(np.nan)
        for vv in builtins.range(Vdim):
            i = np.argsort(values[vv])
            result[vv, binnumbers[i]] = values[vv, i]
    elif callable(statistic):
        with np.errstate(invalid='ignore'), suppress_warnings() as sup:
            sup.filter(RuntimeWarning)
            try:
                null = statistic([])
            except Exception:
                null = np.nan
        if np.iscomplexobj(null):
            result = result.astype(np.complex128)
        result.fill(null)
        try:
            _calc_binned_statistic(
                Vdim, binnumbers, result, values, statistic, values_err=values_err
            )
        except ValueError:
            result = result.astype(np.complex128)
            _calc_binned_statistic(
                Vdim, binnumbers, result, values, statistic, values_err=values_err
            )

    # Shape into a proper matrix
    result = result.reshape(np.append(Vdim, nbin))

    # Remove outliers (indices 0 and -1 for each bin-dimension).
    core = tuple([slice(None)] + Ndim * [slice(1, -1)])
    result = result[core]

    # Unravel binnumbers into an ndarray, each row the bins for each dimension
    if expand_binnumbers and Ndim > 1:
        binnumbers = np.asarray(np.unravel_index(binnumbers, nbin))

    if np.any(result.shape[1:] != nbin - 2):
        raise RuntimeError('Internal Shape Error')

    # Reshape to have output (`result`) match input (`values`) shape
    result = result.reshape(input_shape[:-1] + list(nbin-2))

    return BinnedStatisticddResult(result, edges, binnumbers)


def bin_dd(sample, bins=10, range=None, expand_binnumbers=True, nan_policy='raise'):
    """
    Bins N-dimensional data. Arguments have the same meaning as in scipy.stats.binned_statistic_dd,
    except that here expand_binnumbers=True by default, and that if expand_binnumbers=True and N = 1,
    then binnumbers is 2D instead of 1D as in scipy.stats.binned_statistic_dd.
    nan_policy can be 'raise' or 'omit'. 
    If nan_policy='raise', then ValueError is raised if sample contains any NaNs (this is slightly different
    from the default behavior of scipy.stats.binned_statistic_dd).
    If nan_policy='omit', NaNs are sorted into the rightmost bin.
    """
    
    if nan_policy not in ['raise', 'omit']:
        raise ValueError(f"nan_policy must be 'raise' or 'omit', but {nan_policy=}.")
    
    try:
        bins = index(bins)
    except TypeError:
        # bins is not an integer
        pass
    # If bins was an integer-like object, now it is an actual Python int.

    # NOTE: for _bin_edges(), see e.g. gh-11365
    if nan_policy == 'raise' and not np.isfinite(sample).all():
        raise ValueError('%r contains non-finite values.' % (sample,))

    # `Ndim` is the number of dimensions (e.g. `2` for `binned_statistic_2d`)
    # `Dlen` is the length of elements along each dimension.
    # This code is based on np.histogramdd
    try:
        # `sample` is an ND-array.
        Dlen, Ndim = sample.shape
    except (AttributeError, ValueError):
        # `sample` is a sequence of 1D arrays.
        sample = np.atleast_2d(sample).T
        Dlen, Ndim = sample.shape
        
    try:
        M = len(bins)
        if M != Ndim:
            raise AttributeError('The dimension of bins must be equal '
                                 'to the dimension of the sample x.')
    except TypeError:
        bins = Ndim * [bins]
        
    nbin, edges, dedges = _bin_edges(sample, bins, range)
    binnumbers = _bin_numbers(sample, nbin, edges, dedges, expand_binnumbers=expand_binnumbers)
        
    centers = [np.array([np.nan] + list(0.5*(e[:-1]+e[1:])) + [np.nan]) for e in edges]
    return binnumbers, centers, edges

def bin(sample, bins=10, range=None, nan_policy='raise'):
    try:
        N = len(bins)
    except TypeError:
        N = 1

    if N != 1:
        bins = [np.asarray(bins, float)]

    if range is not None:
        if len(range) == 2:
            range = [range]
            
    bin_nums, centers, edges = bin_dd(sample, bins=bins, range=range, nan_policy=nan_policy)
    return bin_nums[0], centers[0], edges[0]

def binned_mean_dd(x, y, yerr=None, weighted=False, nanstats=True, **kwargs):
    nan = 'nan' if nanstats else ''
    weighted = 'weighted' if weighted else ''
    
    if yerr is not None:
        statistics = {
            'y': lambda y, yerr: getattr(npstats, f'{nan}{weighted}meanerr')(y, yerr)[0],
            'yerr': lambda y, yerr: getattr(npstats, f'{nan}{weighted}meanerr')(y, yerr)[1]
        }
    else:
        if weighted:
            raise ValueError("weighted is True, but yerr is not provided")
            
        statistics = {
            'y': lambda y: getattr(npstats, f'{nan}mean')(y),
            'yerr': lambda y: getattr(npstats, f'{nan}sem')(y),
        }
    
    result = {}
    
    out = binned_statistic_dd(x, y, values_err=yerr, statistic=statistics['y'], **kwargs)
    
    bin_edges = out.bin_edges
    bin_centers = [0.5*(edges[:-1] + edges[1:]) for edges in bin_edges]
    grid = np.stack(np.meshgrid(*bin_centers, indexing='ij'), axis=-1)
    result['x'] = grid
    result['y'] = out.statistic
    
    out = binned_statistic_dd(x, y, values_err=yerr, statistic=statistics['yerr'], binned_statistic_result=out, **kwargs)
    result['yerr'] = out.statistic
    
    return result

def binned_mean(x, y, yerr=None, bins=10, range=None, **kwargs):
    try:
        N = len(bins)
    except TypeError:
        N = 1

    if N != 1:
        bins = [np.asarray(bins, float)]

    if range is not None:
        if len(range) == 2:
            range = [range]

    result = binned_mean_dd(
        [x], y, yerr=yerr, bins=bins, range=range, **kwargs)
    
    result['x'] = result['x'].squeeze()
    
    return result


import collections
import contextlib
import itertools

import torch

__all__ = ['isconst', 'inv_perm', 'lexsort', 'unique', 'meshgrid', 'use_deterministic_algorithms']

def isconst(x, dim=None, **kwargs):
    x = torch.as_tensor(x)
    
    if dim is None:
        x = x.reshape(-1)
    else:
        if isinstance(dim, int):
            dim = [dim]
        if len(dim) == 0:
            return torch.ones(x.shape, device=x.device, dtype=torch.bool)
        dim = sorted([d % x.ndim for d in dim])[::-1]
        for d in dim:
            x = torch.movedim(x, d,-1)
        x = x.reshape(*x.shape[:-len(dim)],-1)
        
    if x.is_floating_point():
        return torch.isclose(x[...,:-1], x[...,1:], **kwargs).all(dim=-1)
    return (x[...,:-1] == x[...,1:]).all(dim=-1)

def inv_perm(x):
    x_inv = torch.empty_like(x)
    x_inv[x] = torch.arange(len(x), device=x.device)
    return x_inv

def lexsort(keys, dim=-1):
    keys = keys if isinstance(keys, torch.Tensor) else torch.stack(keys)
    if keys.ndim < 2:
        raise ValueError(f"keys must be at least 2 dimensional, but {keys.ndim=}.")
    if len(keys) == 0:
        raise ValueError(f"Must have at least 1 key, but {len(keys)=}.")
    
    # to understand reasoning, see the code at
    # https://github.com/wouterkool/attention-learn-to-route/blob/master/utils/lexsort.py#L39-L48
    idx = keys[0].argsort(dim=dim, stable=True)
    for k in keys[1:]:
        idx = idx.gather(dim, k.gather(dim, idx).argsort(dim=dim, stable=True))
    
    return idx

def _unique_sorted(x, dim=None, return_index=False, return_inverse=False, return_counts=False, first_index=True):
    if dim is None:
        x = x.reshape(-1)
        dim = 0
        
    # reshape to a 2D tensor where we compute unique rows
    x = x.movedim(dim, 0)
    shape = x.shape
    x = x.reshape(shape[0], -1)
    
    sort_idx = lexsort(x.t().flip(0)) # note: as of torch.2.0.0 there is a bug with torch.flip on MPS
    x = x[sort_idx]
    
    if return_index:
        x, inverse, counts = torch.unique_consecutive(x, dim=0, return_inverse=True, return_counts=True)
        tot_counts = torch.cat((counts.new_zeros(1), counts.int().cumsum(dim=0).long()))[:-1] # cast to int32 due to MPS limitation
        index = torch.arange(len(inverse), device=tot_counts.device)[tot_counts]
        
        out = {'x': x, 'index': index}
        if return_inverse:
            out['inverse'] = inverse
        if return_counts:
            out['counts'] = counts

        # the following code doesn't work right now on MPS due to MPS determinism bug
#         ret = torch.unique_consecutive(x, dim=0, return_inverse=True, return_counts=return_counts)

#         x, inverse = ret[0], ret[1]
#         args = (len(inverse)-1, -1, -1) if first_index else (len(inverse),)
#         with use_deterministic_algorithms() if first_index else contextlib.nullcontext():
#             # scatter_ is non-deterministic by default and would result in a random index rather than the first index
#             index = inverse.new_empty(len(x)).scatter_(
#                 0,
#                 inverse.flip(0) if first_index else inverse,
#                 torch.arange(*args, dtype=inverse.dtype, device=inverse.device),
#             )
        
#         # gather results into a dictionary
#         out = {'x': x, 'index': index}
#         if return_inverse:
#             out['inverse'] = inverse
#         if return_counts:
#             out['counts'] = ret[2]

    else:
        ret = torch.unique_consecutive(x, dim=0, return_inverse=return_inverse, return_counts=return_counts)

        # gather results into a dictionary
        ret = list(ret) if isinstance(ret, tuple) else [ret]
        out = {'x': ret.pop(0)}
        if return_inverse:
            out['inverse'] = ret.pop(0)
        if return_counts:
            out['counts'] = ret.pop(0)
            
    # correct for the fact that we reshaped and sorted before computing unqiue
    out['x'] = out['x'].reshape(-1, *shape[1:]).movedim(0, dim)
    if return_index or return_inverse:
        if return_index:
            out['index'] = sort_idx[out['index']]
        if return_inverse:
            sort_idx_inv = inv_perm(sort_idx)
            out['inverse'] = out['inverse'][sort_idx_inv]
    
    if len(out) == 1:
        return out['x']
    return tuple(out.values())

def unique(x, dim=None, sorted=True, return_index=False, return_inverse=False, return_counts=False, equal_nan=False, first_index=True):
    """
    Note:
     - When sorted=False, there is no guarantee on what the result order will be.
     - The result of this function when sorted=True, return_index=False is slightly different from numpy
     with return_index=False since numpy does not use stable sort when return_index=False, only when return_index=True.
     - If return_index=True and first_index=True, the returned index is the first index of the unique element. Otherwise,
        the returned index can correspond to any index of the unique element, and the behavior is non-deterministic.
    """
    
    if equal_nan:
        raise NotImplementedError("pytorch currently does not support equal_nan=True")
    
    # unique_dim currently not implemented for MPS, so do _unique_sorted instead
    # which uses unique_consecutive (which has MPS implementation) instead of unique_dim.
    # the reason there is no MPS unqiue_dim support right now is that there is currently 
    # no native MPS support for sorting with custom compare and equal functions, so one 
    # has to write a custom MPS Kernel which means it might take a long while before this gets implemented.
    if sorted or x.device.type == 'mps':
        return _unique_sorted(x, dim=dim, return_index=return_index, return_inverse=return_inverse, return_counts=return_counts)
    
    if dim is None:
        x = x.reshape(-1)
        dim = 0
    
    # reshape to a 2D tensor where we compute unique rows
    x = x.movedim(dim, 0)
    shape = x.shape
    x = x.reshape(shape[0], -1)
    
    # separate rows with nans from rows without nans and only compute unique on rows without nans.
    # This avoids torch.unique bug when dim is not None and input has nan.
    isnan = x.isnan().any(dim=1)
    if not isnan.any():
        x_not_isnan, x_isnan = x, torch.empty((0, x.shape[1]), dtype=x.dtype, device=x.device) # slighty optimization to prevent large copy
    else:
        x_not_isnan, x_isnan = x[~isnan], x[isnan]
    N_not_isnan = len(x_not_isnan)

    if return_index:
        # # code copied from https://github.com/pytorch/pytorch/issues/36748#issuecomment-1474368922
        # x_not_isnan, inverse, counts = torch.unique(x_not_isnan, dim=0, return_inverse=True, return_counts=True)
        # inv_sort_idx = inverse.argsort(stable=True)
        # tot_counts = torch.cat((counts.new_zeros(1), counts.cumsum(dim=0)))[:-1]
        # index = inv_sort_idx[tot_counts]

        # out = {'x': x_not_isnan, 'index': index}
        # if return_inverse:
        #     out['inverse'] = inverse
        # if return_counts:
        #     out['counts'] = counts

        ret = torch.unique(x_not_isnan, dim=0, return_inverse=True, return_counts=return_counts)

        x_not_isnan, inverse = ret[0], ret[1]
        args = (len(inverse)-1, -1, -1) if first_index else (len(inverse),)
        with use_deterministic_algorithms() if first_index else contextlib.nullcontext():
            # scatter_ is non-deterministic by default and would result in a random index rather than the first index
            index = inverse.new_empty(len(x_not_isnan)).scatter_(
                0,
                inverse.flip(0) if first_index else inverse,
                torch.arange(*args, dtype=inverse.dtype, device=inverse.device),
            )

        # gather results into a dictionary
        out = {'x': x_not_isnan, 'index': index}
        if return_inverse:
            out['inverse'] = inverse
        if return_counts:
            out['counts'] = ret[2]

    else:
        ret = torch.unique(x_not_isnan, dim=0, return_inverse=return_inverse, return_counts=return_counts)
    
        # gather results into a dictionary
        ret = list(ret) if isinstance(ret, tuple) else [ret]
        out = {'x': ret.pop(0)}
        if return_inverse:
            out['inverse'] = ret.pop(0)
        if return_counts:
            out['counts'] = ret.pop(0)
        
    # append nans to the end of the unique tensor and handle index, inverse, and counts accordingly
    N_unique_not_isnan = len(out['x'])
    out['x'] = torch.cat([out['x'], x_isnan])
    if return_index:
        out['index'] = torch.cat([out['index'], torch.arange(len(x_isnan), device=x.device) + N_not_isnan])
    if return_inverse:
        out['inverse'] = torch.cat([out['inverse'], torch.arange(len(x_isnan), device=x.device) + N_unique_not_isnan])
    if return_counts:
        out['counts'] = torch.cat([out['counts'], out['counts'].new_ones(len(x_isnan))])
    
    # correct for the fact that we reshaped and modified element order before computing unqiue
    out['x'] = out['x'].reshape(-1, *shape[1:]).movedim(0, dim)
    if return_index or return_inverse:
        sort_idx = torch.cat(((~isnan).nonzero(), isnan.nonzero())).squeeze()
        if return_index:
            out['index'] = sort_idx[out['index']]
        if return_inverse:
            sort_idx_inv = inv_perm(sort_idx)
            out['inverse'] = out['inverse'][sort_idx_inv]
    
    if len(out) == 1:
        return out['x']
    return tuple(out.values())

def meshgrid(*tensors, indexing='ij', ndims=None):
    """
    The most general version of torch.meshgrid. Meshes together tensors of shapes
    (n_1_1,...,n_1_{ndims_1},*_1), (n_2_1,...,n_2_{ndims_2},*_2), ..., (n_P_1,...,n_P_{ndims_P},*_P).
    indexing must be either 'ij' or 'xy'. Defaults to 'ij' like pytorch but unlike numpy.
    If ndims is None, meshes over all dimensions. ndims can be an int or an Iterable.
    Negative ndim has the same semantic meaning as slicing with a negative index.
    
    If indexing == 'ij', returns tensors of shapes 
    (n_1_1,...,n_1_{ndims_1},n_2_1,...,n_2_{ndims_2},...,n_P_1,...,n_P_{ndims_P},*_1),
    (n_1_1,...,n_1_{ndims_1},n_2_1,...,n_2_{ndims_2},...,n_P_1,...,n_P_{ndims_P},*_2),
    ...
    (n_1_1,...,n_1_{ndims_1},n_2_1,...,n_2_{ndims_2},...,n_P_1,...,n_P_{ndims_P},*_P)
    
    Otherwise, returns tensors of shapes
    (n_2_1,...,n_2_{ndims_2},n_1_1,...,n_1_{ndims_1},...,n_P_1,...,n_P_{ndims_P},*_1),
    (n_2_1,...,n_2_{ndims_2},n_1_1,...,n_1_{ndims_1},...,n_P_1,...,n_P_{ndims_P},*_2),
    ...
    (n_2_1,...,n_2_{ndims_2},n_1_1,...,n_1_{ndims_1},...,n_P_1,...,n_P_{ndims_P},*_P)
    
    IMPORTANT: Data is NOT copied just like pytorch, but unlike numpy which copies by default.
    """
    if not indexing in {'ij', 'xy'}:
        raise ValueError(f"indexing must 'ij' or 'xy', but got {indexing}.")
    
    if not isinstance(ndims, collections.abc.Iterable):
        ndims = [ndims] * len(tensors)
    elif iter(ndims) is ndims:
        ndims = list(ndims)
        
    if not all(isinstance(ndim, int) or ndim is None for ndim in ndims):
        raise TypeError(f"ndim must be None or an int, but got {ndims}.")

    if len(ndims) != len(tensors):
        raise ValueError(
            f"""ndims must have same number of elements as input tensors,
            but got {len(ndims)=} and {len(tensors)=}."""
        )
        
    if not all(
        abs(ndim) <= tensor.ndim
        for tensor, ndim in zip(tensors, ndims) if isinstance(ndim, int)
    ):
        raise ValueError(
            f"""abs(ndim) cannot be greater than the corresponding tensor.ndim, but
            {ndims=} while tensor ndims={[tensor.ndim for tensor in tensors]}."""
        )
    
    cumndims = list(itertools.accumulate((
        tensor.ndim if ndim is None else tensor.ndim + ndim if ndim < 0 else ndim
        for tensor, ndim in zip(tensors, ndims)
    )))
    pre_cumndims = [0] + cumndims[:-1]
    post_cumndims = (cumndims[-1] - cumndim for cumndim in cumndims)
    pre_shapes, post_shapes = zip(*(
        (tensor.shape[:ndim], tuple() if ndim is None else tensor.shape[ndim:])
        for tensor, ndim in zip(tensors, ndims)
    ))
    shapes = (
        (1,) * pre_cumndim + pre_shape + (1,) * post_cumndim + post_shape
        for pre_shape, post_shape, pre_cumndim, post_cumndim
        in zip(pre_shapes, post_shapes, pre_cumndims, post_cumndims)
    )
    shared_shape = tuple(itertools.chain(*pre_shapes))
    tensors = (
        tensor.reshape(shape).broadcast_to(shared_shape + post_shape)
        for tensor, shape, post_shape in zip(tensors, shapes, post_shapes)
    )
    
    if indexing == 'ij':
        return list(tensors)
        
    return [
        tensor.movedim(
            tuple(range(cumndims[0], cumndims[1])),
            tuple(range(cumndims[1] - cumndims[0]))
        )
        for tensor in tensors
    ]

@contextlib.contextmanager
def use_deterministic_algorithms():
    """
    Context manager to use deterministic algorithms.
    """
    is_deterministic = torch.are_deterministic_algorithms_enabled()
    try:
        torch.use_deterministic_algorithms(True)
        yield
    finally:
        torch.use_deterministic_algorithms(is_deterministic)

##########################################
### Implemented but untested functions ###
##########################################

# import numpy as np

# def ravel_multi_index(multi_index, dims):
#     """
#     Similar to np.ravel_multi_index
#     """
#     multi_index = torch.stack(multi_index) if not isinstance(multi_index, torch.Tensor) else multi_index
#     dims = torch.as_tensor(dims, device=multi_index.device)
    
#     if len(multi_index) != len(dims):
#         raise ValueError(f"multi_index and dims must have same length, but {len(multi_index)=} and {len(dims)=}.")
    
#     if torch.is_floating_point(multi_index):
#         raise TypeError(f"multi_index must be integer dtype, but {multi_index.dtype=}.")
        
#     if (multi_index.min(dim=1).values < 0).any():
#         raise ValueError(f"multi_index must be non-negative, but {multi_index.min(dim=1).values=}.")
        
#     if (multi_index.max(dim=1).values >= dims).any():
#         raise ValueError(f"multi_index must be less than dims along each dimension, but {multi_index.max(dim=1).values=} and {dims=}.")
        
#     multipliers = np.cumprod((dims[1:].tolist() + [1])[::-1])[::-1]
#     return torch.stack([index * multiplier for index, multiplier in zip(multi_index, multipliers)], dim=0).sum(dim=0)

# def unravel_index(indices, shape, *, as_tuple=True):
#     r"""
#     Modified from https://github.com/pytorch/pytorch/pull/66687/files
    
#     Converts a `Tensor` of flat indices into a `Tensor` of coordinates for the given target shape.
#     Args:
#         indices: An integral `Tensor` containing flattened indices of a `Tensor` of dimension `shape`.
#         shape: The shape (can be an `int`, a `Sequence` or a `Tensor`) of the `Tensor` for which
#                the flattened `indices` are unraveled.
#     Keyword Args:
#         as_tuple: A boolean value, which if `True` will return the result as tuple of Tensors,
#                   else a `Tensor` will be returned. Default: `True`
#     Returns:
#         unraveled coordinates from the given `indices` and `shape`. See description of `as_tuple` for
#         returning a `tuple`.
#     .. note:: The default behaviour of this function is analogous to
#               `numpy.unravel_index <https://numpy.org/doc/stable/reference/generated/numpy.unravel_index.html>`_.
#     Example::
#         >>> indices = torch.tensor([22, 41, 37])
#         >>> shape = (7, 6)
#         >>> torch.unravel_index(indices, shape)
#         (tensor([3, 6, 6]), tensor([4, 5, 1]))
#         >>> torch.unravel_index(indices, shape, as_tuple=False)
#         tensor([[3, 4],
#                 [6, 5],
#                 [6, 1]])
#         >>> indices = torch.tensor([3, 10, 12])
#         >>> shape_ = (4, 2, 3)
#         >>> torch.unravel_index(indices, shape_)
#         (tensor([0, 1, 2]), tensor([1, 1, 0]), tensor([0, 1, 0]))
#         >>> torch.unravel_index(indices, shape_, as_tuple=False)
#         tensor([[0, 1, 0],
#                 [1, 1, 1],
#                 [2, 0, 0]])
#     """
#     def _helper_type_check(inp, name):
#         # `indices` is expected to be a tensor, while `shape` can be a sequence/int/tensor
#         if name == "shape" and isinstance(inp, collections.abc.Sequence):
#             for dim in inp:
#                 if not isinstance(dim, int):
#                     raise TypeError("Expected shape to have only integral elements.")
#                 if dim < 0:
#                     raise ValueError("Negative values in shape are not allowed.")
#         elif name == "shape" and isinstance(inp, int):
#             if inp < 0:
#                 raise ValueError("Negative values in shape are not allowed.")
#         elif isinstance(inp, torch.Tensor):
#             if torch.is_floating_point(inp):
#                 raise TypeError(f"Expected {name} to be an integral tensor, but found dtype: {inp.dtype}")
#             if torch.any(inp < 0):
#                 raise ValueError(f"Negative values in {name} are not allowed.")
#         else:
#             allowed_types = "Sequence/Scalar (int)/Tensor" if name == "shape" else "Tensor"
#             msg = f"{name} should either be a {allowed_types}, but found {type(inp)}"
#             raise TypeError(msg)

#     _helper_type_check(indices, "indices")
#     _helper_type_check(shape, "shape")

#     # Convert to a tensor, with the same properties as that of indices
#     if isinstance(shape, collections.abc.Sequence):
#         shape_tensor = indices.new_tensor(shape)
#     elif isinstance(shape, int) or (isinstance(shape, Tensor) and shape.ndim == 0):
#         shape_tensor = indices.new_tensor((shape,))
#     else:
#         shape_tensor = shape

#     # By this time, shape tensor will have dim = 1 if it was passed as scalar (see if-elif above)
#     assert shape_tensor.ndim == 1, "Expected dimension of shape tensor to be <= 1, "
#     f"but got the tensor with dim: {shape_tensor.ndim}."

#     # In case no indices passed, return an empty tensor with number of elements = shape.numel()
#     if indices.numel() == 0:
#         # If both indices.numel() == 0 and shape.numel() == 0, short-circuit to return itself
#         shape_numel = shape_tensor.numel()
#         if shape_numel == 0:
#             raise ValueError("Got indices and shape as empty tensors, expected non-empty tensors.")
#         else:
#             output = [indices.new_tensor([]) for _ in range(shape_numel)]
#             return tuple(output) if as_tuple else torch.stack(output, dim=1)

#     if torch.max(indices) >= torch.prod(shape_tensor):
#         raise ValueError("Target shape should cover all source indices.")

#     coefs = shape_tensor[1:].flipud().cumprod(dim=0).flipud()
#     coefs = torch.cat((coefs, coefs.new_tensor((1,))), dim=0)
#     coords = torch.div(indices[..., None], coefs, rounding_mode='trunc') % shape_tensor

#     if as_tuple:
#         return tuple(coords[..., i] for i in range(coords.size(-1)))
#     return coords